; -----------------------------------------------------------------------------
; "Agile" integrated RCS+ZX7 decoder by Einar Saukas (150 bytes)
; -----------------------------------------------------------------------------
; Parameters:
;   HL: source address (compressed data)
;   DE: destination address (decompressing)
; -----------------------------------------------------------------------------

dzx7:
        ld      a, $80
copy_byte_loop_rcs:
        ex      af, af'
        call    copy_byte               ; copy literal byte
main_loop_rcs:
        ex      af, af'
        add     a, a                    ; check next bit
        call    z, load_bits            ; no more bits left?
        jr      nc, copy_byte_loop_rcs  ; next bit indicates either literal or sequence
        jr      length_size_start

copy_byte_loop:
        ldi                             ; copy literal byte
main_loop:
        add     a, a                    ; check next bit
        call    z, load_bits            ; no more bits left?
        jr      nc, copy_byte_loop      ; next bit indicates either literal or sequence

length_size_start:
; determine number of bits used for length (Elias gamma coding)
        push    de
        ld      bc, 1
        ld      d, b
length_size_loop:
        inc     d
        add     a, a                    ; check next bit
        call    z, load_bits            ; no more bits left?
        jr      nc, length_size_loop
        jp      length_value_start

; determine length
length_value_loop:
        add     a, a                    ; check next bit
        call    z, load_bits            ; no more bits left?
        rl      c
        rl      b
        jr      c, exit                 ; check end marker
length_value_start:
        dec     d
        jr      nz, length_value_loop
        inc     bc                      ; adjust length

; determine offset
        ld      e, (hl)                 ; load offset flag (1 bit) + offset value (7 bits)
        inc     hl
        defb    $cb, $33                ; opcode for undocumented instruction "SLL E" aka "SLS E"
        jr      nc, offset_end          ; if offset flag is set, load 4 extra bits
        add     a, a                    ; check next bit
        call    z, load_bits            ; no more bits left?
        rl      d                       ; insert first bit into D
        add     a, a                    ; check next bit
        call    z, load_bits            ; no more bits left?
        rl      d                       ; insert second bit into D
        add     a, a                    ; check next bit
        call    z, load_bits            ; no more bits left?
        rl      d                       ; insert third bit into D
        add     a, a                    ; check next bit
        call    z, load_bits            ; no more bits left?
        ccf
        jr      c, offset_end
        inc     d                       ; equivalent to adding 128 to DE
offset_end:
        rr      e                       ; insert inverted fourth bit into E

; copy previous sequence
        ex      (sp), hl                ; store source, restore destination
        push    hl                      ; store destination
        sbc     hl, de                  ; HL = destination - offset - 1
        pop     de                      ; DE = destination
        ex      af, af'
        ld      a, h                    ; A = 010RRccc
        cp      $58
        jr      c, copy_bytes
        ldir
        pop     hl                      ; restore source address (compressed data)
        ex      af, af'
        jp      main_loop

copy_bytes:
        push    hl
        ex      de, hl
        call    convert
        ex      de, hl
        call    copy_byte
        pop     hl
        inc     hl
        jp      pe, copy_bytes
        pop     hl                      ; restore source address (compressed data)
        jr      main_loop_rcs

load_bits:
        ld      a, (hl)                 ; load another group of 8 bits
        inc     hl
        rla
        ret

copy_byte:
        push    de
        call    convert
        ldi
exit:
        pop     de
        inc     de
        ret

; Convert an RCS address 010RRccc ccrrrppp to screen address 010RRppp rrrccccc
; (Note: replace both EX AF,AF' with PUSH AF/POP AF if you want to preserve AF')
convert:
        ld      a, d                    ; A = 010RRccc
        cp      $58
        ret     nc
        xor     e
        and     $f8
        xor     e                       ; A = 010RRppp
        push    af
        xor     d
        xor     e                       ; A = ccrrrccc
        rlca
        rlca                            ; A = rrrccccc
        pop     de                      ; D = 010RRppp
        ld      e, a                    ; E = rrrccccc
        ret

; -----------------------------------------------------------------------------
